from datetime import datetime, timedelta, timezone
from unittest.mock import call, patch

from django.test import override_settings
from django.urls import reverse
from rest_framework import status

from arkindex.documents.models import Corpus, CorpusExportState
from arkindex.project.tests import FixtureAPITestCase
from arkindex.users.models import Role


class TestExport(FixtureAPITestCase):

    @patch("arkindex.project.triggers.export.export_corpus.delay")
    @override_settings(EXPORT_TTL_SECONDS=420)
    def test_start(self, delay_mock):
        self.client.force_login(self.superuser)
        response = self.client.post(reverse("api:corpus-export", kwargs={"pk": self.corpus.id}))
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)

        export = self.corpus.exports.get()
        self.assertDictEqual(response.json(), {
            "id": str(export.id),
            "created": export.created.isoformat().replace("+00:00", "Z"),
            "updated": export.updated.isoformat().replace("+00:00", "Z"),
            "user": {
                "id": self.superuser.id,
                "display_name": self.superuser.display_name,
                "email": self.superuser.email,
            },
            "corpus_id": str(self.corpus.id),
            "state": CorpusExportState.Created.value,
        })

        self.assertEqual(delay_mock.call_count, 1)
        self.assertEqual(delay_mock.call_args, call(
            corpus_export=export,
            user_id=self.superuser.id,
            description="Export of corpus Unit Tests"
        ))

    @patch("arkindex.project.triggers.export.export_corpus.delay")
    def test_start_requires_login(self, delay_mock):
        response = self.client.post(reverse("api:corpus-export", kwargs={"pk": self.corpus.id}))
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

        self.assertFalse(self.corpus.exports.exists())
        self.assertFalse(delay_mock.called)

    @patch("arkindex.project.triggers.export.export_corpus.delay")
    def test_start_requires_verified(self, delay_mock):
        self.user.verified_email = False
        self.user.save()
        self.client.force_login(self.user)

        response = self.client.post(reverse("api:corpus-export", kwargs={"pk": self.corpus.id}))
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

        self.assertFalse(self.corpus.exports.exists())
        self.assertFalse(delay_mock.called)

    @patch("arkindex.project.triggers.export.export_corpus.delay")
    @patch("arkindex.project.mixins.has_access", return_value=False)
    def test_start_requires_contributor(self, has_access_mock, delay_mock):
        self.user.rights.update(level=Role.Guest.value)
        self.client.force_login(self.user)

        response = self.client.post(reverse("api:corpus-export", kwargs={"pk": self.corpus.id}))
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

        self.assertEqual(has_access_mock.call_count, 1)
        self.assertEqual(has_access_mock.call_args, call(self.user, self.corpus, Role.Contributor.value, skip_public=False))
        self.assertFalse(self.corpus.exports.exists())
        self.assertFalse(delay_mock.called)

    @patch("arkindex.project.triggers.export.export_corpus.delay")
    def test_start_running(self, delay_mock):
        self.client.force_login(self.superuser)
        self.corpus.exports.create(user=self.user, state=CorpusExportState.Running)

        response = self.client.post(reverse("api:corpus-export", kwargs={"pk": self.corpus.id}))
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertListEqual(response.json(), ["An export is already running for this corpus."])

        self.assertEqual(self.corpus.exports.count(), 1)
        self.assertFalse(delay_mock.called)

    @override_settings(EXPORT_TTL_SECONDS=420)
    @patch("arkindex.project.triggers.export.export_corpus.delay")
    def test_start_recent_export(self, delay_mock):
        self.client.force_login(self.superuser)
        with patch("django.utils.timezone.now") as mock_now:
            mock_now.return_value = datetime.now(timezone.utc) - timedelta(minutes=2)
            self.corpus.exports.create(
                user=self.user,
                state=CorpusExportState.Done,
            )

        response = self.client.post(reverse("api:corpus-export", kwargs={"pk": self.corpus.id}))
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertListEqual(response.json(), ["An export has already been made for this corpus in the last 420 seconds."])

        self.assertEqual(self.corpus.exports.count(), 1)
        self.assertFalse(delay_mock.called)

    def test_list(self):
        export1 = self.corpus.exports.create(user=self.user, state=CorpusExportState.Done)
        export2 = self.corpus.exports.create(user=self.superuser)
        self.client.force_login(self.superuser)

        response = self.client.get(reverse("api:corpus-export", kwargs={"pk": self.corpus.id}))
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertListEqual(response.json()["results"], [
            {
                "id": str(export2.id),
                "created": export2.created.isoformat().replace("+00:00", "Z"),
                "updated": export2.updated.isoformat().replace("+00:00", "Z"),
                "state": CorpusExportState.Created.value,
                "user": {
                    "id": self.superuser.id,
                    "display_name": self.superuser.display_name,
                    "email": self.superuser.email,
                },
                "corpus_id": str(self.corpus.id),
            },
            {
                "id": str(export1.id),
                "created": export1.created.isoformat().replace("+00:00", "Z"),
                "updated": export1.updated.isoformat().replace("+00:00", "Z"),
                "state": CorpusExportState.Done.value,
                "user": {
                    "id": self.user.id,
                    "display_name": self.user.display_name,
                    "email": self.user.email,
                },
                "corpus_id": str(self.corpus.id),
            },
        ])

    def test_list_requires_login(self):
        response = self.client.get(reverse("api:corpus-export", kwargs={"pk": self.corpus.id}))
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_list_requires_verified(self):
        self.user.verified_email = False
        self.user.save()
        self.client.force_login(self.user)
        response = self.client.get(reverse("api:corpus-export", kwargs={"pk": self.corpus.id}))
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    @patch("arkindex.project.mixins.has_access", return_value=False)
    def test_list_requires_guest(self, has_access_mock):
        self.user.rights.all().delete()
        self.corpus.public = False
        self.corpus.save()
        self.client.force_login(self.user)

        response = self.client.get(reverse("api:corpus-export", kwargs={"pk": self.corpus.id}))
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

        self.assertEqual(has_access_mock.call_count, 1)
        self.assertEqual(has_access_mock.call_args, call(self.user, self.corpus, Role.Guest.value, skip_public=False))

    @patch("arkindex.project.aws.s3.meta.client.generate_presigned_url")
    def test_download_export(self, presigned_url_mock):
        presigned_url_mock.return_value = "http://somewhere"
        self.client.force_login(self.superuser)
        export = self.corpus.exports.create(user=self.user, state=CorpusExportState.Done)
        with self.assertNumQueries(3):
            response = self.client.get(reverse("api:manage-export", kwargs={"pk": export.id}))
            self.assertEqual(response.status_code, status.HTTP_302_FOUND)
        self.assertEqual(response.headers["Location"], "http://somewhere")
        self.assertEqual(presigned_url_mock.call_args_list, [
            call(
                "get_object",
                Params={
                    "Bucket": "export",
                    "Key": str(export.id),
                },
            )
        ])

    @patch("arkindex.project.aws.s3.meta.client.generate_presigned_url")
    def test_download_export_public_corpus(self, presigned_url_mock):
        presigned_url_mock.return_value = "http://somewhere"
        self.client.force_login(self.user)
        self.corpus.public = True
        self.corpus.save()
        self.corpus.memberships.filter(user=self.user).delete()
        export = self.corpus.exports.create(user=self.superuser, state=CorpusExportState.Done)

        with self.assertNumQueries(3):
            response = self.client.get(reverse("api:manage-export", kwargs={"pk": export.id}))
            self.assertEqual(response.status_code, status.HTTP_302_FOUND)
        self.assertEqual(response.headers["Location"], "http://somewhere")

    def test_download_export_requires_login(self):
        export = self.corpus.exports.create(user=self.user, state=CorpusExportState.Done)
        response = self.client.get(reverse("api:manage-export", kwargs={"pk": export.id}))
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_download_export_requires_verified(self):
        self.user.verified_email = False
        self.user.save()
        self.client.force_login(self.user)
        export = self.corpus.exports.create(user=self.user, state=CorpusExportState.Done)
        with self.assertNumQueries(2):
            response = self.client.get(reverse("api:manage-export", kwargs={"pk": export.id}))
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    @patch("arkindex.users.managers.BaseACLManager.filter_rights", return_value=Corpus.objects.none())
    def test_download_export_requires_guest(self, filter_rights_mock):
        self.user.rights.all().delete()
        self.corpus.public = False
        self.corpus.save()
        self.client.force_login(self.user)
        export = self.corpus.exports.create(user=self.user, state=CorpusExportState.Done)
        with self.assertNumQueries(2):
            response = self.client.get(reverse("api:manage-export", kwargs={"pk": export.id}))
            self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        self.assertEqual(filter_rights_mock.call_count, 1)
        self.assertEqual(filter_rights_mock.call_args, call(self.user, Corpus, Role.Guest.value))

    def test_download_export_not_done(self):
        self.client.force_login(self.superuser)
        for state in (CorpusExportState.Created, CorpusExportState.Running, CorpusExportState.Failed):
            with self.subTest(state=state):
                export = self.corpus.exports.create(user=self.user, state=state)
                with self.assertNumQueries(3):
                    response = self.client.get(reverse("api:manage-export", kwargs={"pk": export.id}))
                    self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

    # DestroyExport

    def test_delete_export_requires_login(self):
        export = self.corpus.exports.create(user=self.user, state=CorpusExportState.Done)
        response = self.client.delete(reverse("api:manage-export", kwargs={"pk": export.id}))
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_delete_export_requires_verified(self):
        self.user.verified_email = False
        self.user.save()
        self.client.force_login(self.user)
        export = self.corpus.exports.create(user=self.user, state=CorpusExportState.Done)
        with self.assertNumQueries(2):
            response = self.client.delete(reverse("api:manage-export", kwargs={"pk": export.id}))
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    @patch("arkindex.users.managers.BaseACLManager.filter_rights", return_value=Corpus.objects.none())
    def test_delete_export_private_corpus(self, filter_rights_mock):
        private_corpus = Corpus.objects.create(name="private")
        self.client.force_login(self.user)
        export = private_corpus.exports.create(user=self.superuser, state=CorpusExportState.Done)
        with self.assertNumQueries(2):
            response = self.client.delete(reverse("api:manage-export", kwargs={"pk": export.id}))
            self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        self.assertEqual(filter_rights_mock.call_count, 1)
        self.assertEqual(filter_rights_mock.call_args, call(self.user, Corpus, Role.Guest.value))

    def test_delete_export_wrong_state(self):
        self.client.force_login(self.superuser)
        for state in (CorpusExportState.Created, CorpusExportState.Failed):
            with self.subTest(state=state):
                export = self.corpus.exports.create(user=self.user, state=state)
                with self.assertNumQueries(3):
                    response = self.client.delete(reverse("api:manage-export", kwargs={"pk": export.id}))
                    self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

    @override_settings(EXPORT_TTL_SECONDS=420)
    def test_delete_export_running(self):
        self.client.force_login(self.superuser)
        with patch("django.utils.timezone.now") as mock_now:
            mock_now.return_value = datetime.now(timezone.utc) - timedelta(weeks=3)
            export = self.corpus.exports.create(user=self.user, state=CorpusExportState.Running)
        self.corpus.exports.filter(id=export.id).update(updated=datetime.now(timezone.utc) - timedelta(minutes=2))
        with self.assertNumQueries(3):
            response = self.client.delete(reverse("api:manage-export", kwargs={"pk": export.id}))
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertEqual(response.json(), ["You cannot delete an export that is still running."])

    @override_settings(EXPORT_TTL_SECONDS=420)
    def test_delete_export_running_expired(self):
        self.client.force_login(self.superuser)
        with patch("django.utils.timezone.now") as mock_now:
            mock_now.return_value = datetime.now(timezone.utc) - timedelta(weeks=3)
            export = self.corpus.exports.create(user=self.user, state=CorpusExportState.Running)
        self.corpus.exports.filter(id=export.id).update(updated=datetime.now(timezone.utc) - timedelta(minutes=10))
        with self.assertNumQueries(4):
            response = self.client.delete(reverse("api:manage-export", kwargs={"pk": export.id}))
            self.assertEqual(response.status_code, status.HTTP_204_NO_CONTENT)
        assert not self.corpus.exports.exists()

    @patch("arkindex.users.utils.get_max_level", return_value=Role.Contributor.value)
    def test_delete_export_requires_admin(self, get_max_level_mock):
        self.corpus.memberships.filter(user=self.user).update(level=Role.Contributor.value)
        self.client.force_login(self.user)
        export = self.corpus.exports.create(user=self.superuser, state=CorpusExportState.Done)

        with self.assertNumQueries(3):
            response = self.client.delete(reverse("api:manage-export", kwargs={"pk": export.id}))
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

        self.assertDictEqual(response.json(), {"detail": "You do not have sufficient rights to delete this export."})

        self.assertEqual(get_max_level_mock.call_count, 1)
        self.assertEqual(get_max_level_mock.call_args, call(self.user, self.corpus))

    @patch("arkindex.users.utils.get_max_level", return_value=Role.Guest.value)
    def test_delete_export_creator_requires_contributor(self, get_max_level_mock):
        self.client.force_login(self.user)
        export = self.corpus.exports.create(user=self.user, state=CorpusExportState.Done)

        with self.assertNumQueries(3):
            response = self.client.delete(reverse("api:manage-export", kwargs={"pk": export.id}))
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

        self.assertDictEqual(response.json(), {"detail": "You do not have sufficient rights to delete this export."})

        self.assertEqual(get_max_level_mock.call_count, 1)
        self.assertEqual(get_max_level_mock.call_args, call(self.user, self.corpus))

    @patch("arkindex.users.utils.get_max_level", return_value=Role.Contributor.value)
    def test_delete_export_creator_contributor(self, get_max_level_mock):
        self.corpus.memberships.filter(user=self.user).update(level=Role.Contributor.value)
        self.client.force_login(self.user)
        export = self.corpus.exports.create(user=self.user, state=CorpusExportState.Done)
        with self.assertNumQueries(4):
            response = self.client.delete(reverse("api:manage-export", kwargs={"pk": export.id}))
            self.assertEqual(response.status_code, status.HTTP_204_NO_CONTENT)
        assert not self.corpus.exports.exists()

        self.assertEqual(get_max_level_mock.call_count, 1)
        self.assertEqual(get_max_level_mock.call_args, call(self.user, self.corpus))

    def test_delete_export_corpus_admin(self):
        self.client.force_login(self.user)
        export = self.corpus.exports.create(user=self.superuser, state=CorpusExportState.Done)
        with self.assertNumQueries(4):
            response = self.client.delete(reverse("api:manage-export", kwargs={"pk": export.id}))
            self.assertEqual(response.status_code, status.HTTP_204_NO_CONTENT)
        assert not self.corpus.exports.exists()
